#include "rdt_wt.h"

/******************************************************************************************
Creates an spatial wavelt tree to index points
*******************************************************************************************/
rdt_wt::rdt_wt(INPUT_MBR *input_data, uint input_size, uint blockSize, char *encoder, int scaleFactor) {
  
  register uint i,j;
  
  //aux structures to map a GIS to a matrix
  number_of_elements = input_size;
  double* ax = (double *) malloc(number_of_elements * sizeof(double));
  double* ay = (double *) malloc(number_of_elements * sizeof(double));
  COORD* cx = (COORD *) malloc(number_of_elements * sizeof(COORD));
  COORD* cy = (COORD *) malloc(number_of_elements * sizeof(COORD));
  double* cxDouble = (double *) malloc(number_of_elements * sizeof(double));
  double* cyDouble = (double *) malloc(number_of_elements * sizeof(double));
  
  //translate input in local structures
  for(i = 0; i < number_of_elements; i++) {
    cx[i].id = cy[i].id = input_data[i].id;
    ax[i] = input_data[i].mbr.point1.lon;
    ay[i] = input_data[i].mbr.point1.lat;
    cx[i].coord = input_data[i].mbr.point2.lon;
    cy[i].coord = input_data[i].mbr.point2.lat;
    cxDouble[i] = input_data[i].mbr.point2.lon;
    cyDouble[i] = input_data[i].mbr.point2.lat;
  }
  
  //order aux structures
  qsort(ax, number_of_elements, sizeof(double), cmpdouble);
  qsort(ay, number_of_elements, sizeof(double), cmpdouble);
  qsort(cxDouble, number_of_elements, sizeof(double), cmpdouble);
  qsort(cyDouble, number_of_elements, sizeof(double), cmpdouble);
  qsort(cx, number_of_elements, sizeof(COORD), cmpcoords);
  qsort(cy, number_of_elements, sizeof(COORD), cmpcoords);
  
  idsCX = (uint *) malloc(number_of_elements * sizeof(uint));
  idsCY = (uint *) malloc(number_of_elements * sizeof(uint));
  
  for(i = 0; i < number_of_elements; i++) {
    idsCX[i] = cx[i].id;
    idsCY[i] = cy[i].id;
  }
  
  if(!strcmp(encoder,PLAIN)) {
    fax = new PlainArray(ax, number_of_elements);
    fay = new PlainArray(ay, number_of_elements);
    fcx = new PlainArray(cxDouble, number_of_elements);
    fcy = new PlainArray(cyDouble, number_of_elements);
  } else {
    fax =  new EncodedArray(ax, number_of_elements, blockSize, encoder, scaleFactor);
    fay =  new EncodedArray(ay, number_of_elements, blockSize, encoder, scaleFactor);
    fcx =  new EncodedArray(cxDouble, number_of_elements, blockSize, encoder, scaleFactor);
    fcy =  new EncodedArray(cyDouble, number_of_elements, blockSize, encoder, scaleFactor);
  }
  
  free(ax);
  free(cxDouble);
  free(ay);
  free(cyDouble);
  
  //allocate wt's
  wtLevelSize = nearest2Pow(number_of_elements);//size of the level in the equivalent complete and balanced but ficticious WT
  wtLevels = logb2(wtLevelSize);//number of levels
  wtSize = number_of_elements * wtLevels;//size of the wt
  uint wtAllocSize = wtSize / 32;//real size of the wt
  if((wtSize % 32) != 0) {
    wtAllocSize ++;
  }
  lonwt = (uint *)malloc(wtAllocSize*sizeof(uint));
  latwt = (uint *)malloc(wtAllocSize*sizeof(uint));
  for(i = 0; i < wtAllocSize; i++) {
    lonwt[i] = 0;
    latwt[i] = 0;
  }
  
  //---------- initialize lonwt -------------------
  //order input as ax
  qsort(input_data, number_of_elements, sizeof(INPUT_MBR), cmpmbrbyax);
  
  //first level
  int* actualPermut = (int *) malloc(number_of_elements * sizeof(int));
  for(i = 0; i < number_of_elements; i++) {
    //input data must be ordered in the order of ax
    actualPermut[i] = translatePermut(input_data, 1, i, cx);
    if(actualPermut[i] >= (int)(wtLevelSize / 2)) {
      bitset(lonwt, i);
    } else {
      bitclean(lonwt, i);
    }
  }
  //end first level
  
  //remaining levels
  uint *chunkLevelCounters = (uint*) malloc((1 << (wtLevels - 1)) * sizeof(uint));
  for(i = 1; i < wtLevels; i++) {
    uint number_of_chunks = 1 << i;
    uint chunkSize = wtLevelSize / number_of_chunks;
    uint levelOffset = i * number_of_elements;
    for(j = 0; j < number_of_chunks; j++) {
      chunkLevelCounters[j] = 0;
    }
    for(j = 0; j < number_of_elements; j++) {
      int actualElement = actualPermut[j];
      uint chunk = getChunk(actualElement, chunkSize);
      uint chunkOffset = chunk * chunkSize;
      uint pos = levelOffset + chunkOffset + chunkLevelCounters[chunk];
      if(actualElement >= (int)(chunkOffset + chunkSize / 2)) {
	bitset(lonwt, pos);
      } else {
	bitclean(lonwt, pos);
      }
      chunkLevelCounters[chunk]++;
    }
  }
  //end remaining levels
  
  #if BITMAP_LEVELS == 1
  lonwtBitmap = createBitmap(lonwt, wtSize,BITMAP_FACTOR);
  #else
  lonwtBitmap = createBitmap(lonwt, wtSize);
  #endif
  
  //---------- end initialize lonwt -------------------
  
  //---------- initialize latwt -------------------
  //order input as ay
  qsort(input_data, number_of_elements, sizeof(INPUT_MBR), cmpmbrbyay);
  
  //first level
  for(i = 0; i < number_of_elements; i++) {
    //input data must be ordered in the order of ax
    actualPermut[i] = translatePermut(input_data, 2, i, cy);
    if(actualPermut[i] >= (int)(wtLevelSize / 2)) {
      bitset(latwt, i);
    } else {
      bitclean(latwt, i);
    }
  }
  //end first level
  
  //remaining levels
  for(i = 1; i < wtLevels; i++) {
    uint number_of_chunks = 1 << i;
    uint chunkSize = wtLevelSize / number_of_chunks;
    uint levelOffset = i * number_of_elements;
    for(j = 0; j < number_of_chunks; j++) {
      chunkLevelCounters[j] = 0;
    }
    for(j = 0; j < number_of_elements; j++) {
      int actualElement = actualPermut[j];
      uint chunk = getChunk(actualElement, chunkSize);
      uint chunkOffset = chunk * chunkSize;
      uint pos = levelOffset + chunkOffset + chunkLevelCounters[chunk];
      if(actualElement >= (int)(chunkOffset + chunkSize / 2)) {
	bitset(latwt, pos);
      } else {
	bitclean(latwt, pos);
      }
      chunkLevelCounters[chunk]++;
    }
  }
  //end remaining levels
  
  free(cx);
  free(cy);

  #if BITMAP_LEVELS == 1
  latwtBitmap = createBitmap(latwt, wtSize,BITMAP_FACTOR);
  #else
  latwtBitmap = createBitmap(latwt, wtSize);
  #endif

  //---------- end initialize lonwt -------------------
  
  free(chunkLevelCounters);
  free(actualPermut);
  free(input_data);
  
  //auxiliary structures
  subChunkUpperLimits = (uint*) malloc(number_of_elements * sizeof(uint));
  newSubChunkUpperLimits = (uint*) malloc(number_of_elements * sizeof(uint));
  candidatePositions = (uint*) malloc(number_of_elements * sizeof(uint));
  
  intersectorSize = INTERSECTOR_SIZE / 32;
  if((wtSize % 32) != 0) {
    intersectorSize ++;
  }
  intersector = (uint *)malloc(intersectorSize*sizeof(uint));
}


//Aux functions
int rdt_wt::translatePermut(INPUT_MBR *mbrsOrderedByCoords, uint type, uint element, COORD *otherCoords) {
  COORD targetCoord;
  targetCoord.id = mbrsOrderedByCoords[element].id;
  if(type == 1) {
    targetCoord.coord = mbrsOrderedByCoords[element].mbr.point2.lon;
  } else {
    targetCoord.coord = mbrsOrderedByCoords[element].mbr.point2.lat;
  }
  COORD *searchResult = (COORD *)bsearch(&targetCoord, otherCoords, number_of_elements, sizeof(COORD), cmpcoords);
  if(searchResult -> id == targetCoord.id) {
    return searchResult - otherCoords;
  } else {
    register uint i;
    for(i = 1; i <= (number_of_elements / 2); i++) {
      COORD *auxSearchResult = searchResult + i;
      if(auxSearchResult < (otherCoords + number_of_elements) && auxSearchResult -> id == targetCoord.id) {
	return auxSearchResult - otherCoords;
      }
      auxSearchResult = searchResult - i;
      if(auxSearchResult >= otherCoords && auxSearchResult -> id == targetCoord.id) {
	return auxSearchResult - otherCoords;
      }
    }
  }
  return -1;
}

int rdt_wt::getChunk(int element, uint chunkSize) {
  return element / chunkSize;
}



/******************************************************************************************
Executes a spatial query (shared with bb_nieves). They differ only in translateQuery
*******************************************************************************************/
int rdt_wt::executeQueryWT(INPUT_MBR *mbr, long *result, uint *resultSize) {
  uint maxax,mincx,maxay,mincy;
  
  clearIntersector();
  if(translateQuery(&mbr -> mbr, &maxax, &mincx, &maxay, &mincy) == -1) {
    return 0;
  }
  //printf("AX to %d; CX from %d; AY to %d; CY from %d\n", maxax, mincx, maxay, mincy);
  
/*  
  //statistics
  candidates1[0] += (1.0*(maxax + 1)/number_of_elements);
  candidates1[1] += (1.0*(maxay + 1)/number_of_elements);
  candidates2[0] += (1.0*(number_of_elements + 1 - mincx)/number_of_elements);
  candidates2[1] += (1.0*(number_of_elements + 1 - mincy)/number_of_elements);
  //
*/

  uint subChunkSize;
  uint *wt;
  bitmap wtBitmap;
  uint minc_;
  uint *ids;
  uint *aux;
  uint newSubChunkSize;
  uint *auxCandidatePositions;
  uint candidatePositionsSize;
  uint newCandidatePositionsSize;	
  uint levelOffset;
  uint chunkSize;
  uint nextLevelChunkSize;
  uint chunk;
  uint nextLevelChunk;
  uint nextLevelChunkLimit;
  uint chunkOffset;
  uint absolutOffset = 0;
  uint relativePos;
  uint absolutPos;
  uint elementsInCompleteNodes;
  uint elementsInNonCompleteNodes;
  uint numberOfPreviousElements=0;
  uint localNumberOfPreviousElements;
  uint subChunkUpperRelativePos;
  uint subChunkUpperAbsolutePos;
  uint value;
  uint subChunkUpperRank;
  uint nextLevelChunkUpperLimit;
  long auxid;
  register uint i;
  register uint j;
  register uint k;
  
  uint inserted = 0;
  
  //---------------------------------------------------- longitudes WT ----------------------------------------------------------------------------------------
  subChunkSize = 1;
  auxCandidatePositions = candidatePositions;
  candidatePositionsSize = 0;
  newCandidatePositionsSize = 0;
  levelOffset = 0;
  chunkSize = wtLevelSize;
  nextLevelChunkSize = wtLevelSize >> 1;
  wt = lonwt;
  wtBitmap = lonwtBitmap;
  minc_ = mincx;
  ids = idsCX;
  subChunkUpperLimits[0] = maxax;
  
  for(i = 0; i < wtLevels; i++) {
    //candidate positions block
    if(candidatePositionsSize != 0) {
      newCandidatePositionsSize = 0;
      auxCandidatePositions = candidatePositions;
      for(j = 0; j < candidatePositionsSize; j++) {
	relativePos = *auxCandidatePositions++;
	chunk = relativePos / chunkSize;
	chunkOffset = chunk * chunkSize;
	absolutOffset = levelOffset + chunkOffset;
	nextLevelChunk = chunk << 1;
	nextLevelChunkLimit = nextLevelChunk * nextLevelChunkSize;
	nextLevelChunkUpperLimit = nextLevelChunkLimit + nextLevelChunkSize;
	//numberOfPreviousElements = rank(wtBitmap, absolutOffset)-bitget(wt,absolutOffset);
	localNumberOfPreviousElements = numberOfPreviousElements + chunk * nextLevelChunkSize;
	absolutPos = levelOffset + relativePos;
	value = bitget(wt, absolutPos);
	if(value) {
	  relativePos = chunkOffset + nextLevelChunkSize + rank(wtBitmap, absolutPos) - localNumberOfPreviousElements - 1;
	  candidatePositions[newCandidatePositionsSize++] = relativePos;
	} else if(minc_ < nextLevelChunkUpperLimit) {
	  relativePos += (localNumberOfPreviousElements- rank(wtBitmap, absolutPos));
	  candidatePositions[newCandidatePositionsSize++] = relativePos;
	}
      }
      candidatePositionsSize = newCandidatePositionsSize;
    }
    
    //subchunks block
    newSubChunkSize = 0;
    for(j = 0; j < subChunkSize; j++) {
      subChunkUpperRelativePos = subChunkUpperLimits[j];
      chunk = subChunkUpperRelativePos / chunkSize;
      chunkOffset = chunk * chunkSize;
      absolutOffset = levelOffset + chunkOffset;
      //numberOfPreviousElements = rank(wtBitmap, absolutOffset)-bitget(wt,absolutOffset);
      localNumberOfPreviousElements = numberOfPreviousElements + chunk * nextLevelChunkSize;
      nextLevelChunk = chunk << 1;
      nextLevelChunkLimit = nextLevelChunk * nextLevelChunkSize;
      nextLevelChunkUpperLimit = nextLevelChunkLimit + nextLevelChunkSize;
      subChunkUpperAbsolutePos = levelOffset + subChunkUpperRelativePos;
      subChunkUpperRank = rank(wtBitmap, subChunkUpperAbsolutePos) - localNumberOfPreviousElements;
      if(minc_ < nextLevelChunkUpperLimit) {
	//branch 0
	subChunkUpperRelativePos = 1 + subChunkUpperRelativePos - subChunkUpperRank;
	if(subChunkUpperRelativePos > nextLevelChunkLimit) {
	  if(subChunkUpperRelativePos == nextLevelChunkUpperLimit && minc_ <= nextLevelChunkLimit) {
	    for(k = 0; k < nextLevelChunkSize; k++) {
	      bitset(intersector, ids[nextLevelChunkLimit + k]);
	    }
	    inserted = 1;
	    //numReportByLevel[0][i] += nextLevelChunkSize;
	    //results[0] += nextLevelChunkSize;
	  } else {
	    subChunkUpperRelativePos--;
	    if(subChunkUpperRelativePos == 0) {
	      candidatePositions[candidatePositionsSize++] = subChunkUpperRelativePos;
	    } else {
	      newSubChunkUpperLimits[newSubChunkSize++] = subChunkUpperRelativePos;
	    }
	  }
	}
      }
      //branch 1 -> always
      subChunkUpperRelativePos = subChunkUpperRank;
      if(subChunkUpperRelativePos > 0) {
	if(minc_ <= nextLevelChunkUpperLimit && subChunkUpperRelativePos == (nextLevelChunkUpperLimit + nextLevelChunkSize)) {
	  for(k = 0; k < nextLevelChunkSize; k++) {
	    bitset(intersector, ids[nextLevelChunkUpperLimit + k]);
	  }
	  inserted = 1;
	  //numReportByLevel[0][i] += nextLevelChunkSize;
	  //results[0] += nextLevelChunkSize;
	} else {
	  subChunkUpperRelativePos--;
	  if(subChunkUpperRelativePos == 0) {
	    //one single element
	    candidatePositions[candidatePositionsSize++] = chunkOffset + nextLevelChunkSize + subChunkUpperRelativePos;
	  } else {
	    newSubChunkUpperLimits[newSubChunkSize++] = chunkOffset + nextLevelChunkSize + subChunkUpperRelativePos;
	  }
	}
      }
      //end branch 1
    }
    subChunkSize = newSubChunkSize;
    
    if(subChunkSize == 0 && candidatePositionsSize == 0) {
      break;
    }

    elementsInCompleteNodes = (number_of_elements/chunkSize)*chunkSize;
    elementsInNonCompleteNodes = number_of_elements - elementsInCompleteNodes;
    numberOfPreviousElements += elementsInCompleteNodes/2;
    if(elementsInNonCompleteNodes>nextLevelChunkSize)
      numberOfPreviousElements += (elementsInNonCompleteNodes-nextLevelChunkSize);

    chunkSize = nextLevelChunkSize;
    nextLevelChunkSize = nextLevelChunkSize >> 1;
    aux = subChunkUpperLimits;
    subChunkUpperLimits = newSubChunkUpperLimits;
    newSubChunkUpperLimits = aux;
    levelOffset += number_of_elements;    
  }
  
  for(j = 0; j < candidatePositionsSize; j++) {
    bitset(intersector, ids[candidatePositions[j]]);
    inserted = 1;
    //numReportByLevel[0][wtLevels-1]++;
    //results[0] ++;;
  }
  
  //---------------------------------------------------- latitudes WT ----------------------------------------------------------------------------------------
  
  if(inserted) {
    subChunkSize = 1;
    subChunkUpperLimits[0] = maxax;
    auxCandidatePositions = candidatePositions;
    candidatePositionsSize = 0;
    newCandidatePositionsSize = 0;
    absolutOffset = 0;
    levelOffset = 0;
    numberOfPreviousElements = 0;
    chunkSize = wtLevelSize;
    nextLevelChunkSize = wtLevelSize >> 1;
    wt = latwt;
    wtBitmap = latwtBitmap;
    minc_ = mincy;
    ids = idsCY;
    subChunkUpperLimits[0] = maxay;
    
    for(i = 0; i < wtLevels; i++) {
      //candidate positions block
      if(candidatePositionsSize != 0) {
	newCandidatePositionsSize = 0;
	auxCandidatePositions = candidatePositions;
	for(j = 0; j < candidatePositionsSize; j++) {
	  relativePos = *auxCandidatePositions++;
	  chunk = relativePos / chunkSize;
	  chunkOffset = chunk * chunkSize;
	  absolutOffset = levelOffset + chunkOffset;
	  nextLevelChunk = 2 * chunk;
	  nextLevelChunkLimit = nextLevelChunk * nextLevelChunkSize;
	  nextLevelChunkUpperLimit = nextLevelChunkLimit + nextLevelChunkSize;
	  //numberOfPreviousElements = rank(wtBitmap, absolutOffset)-bitget(wt,absolutOffset);
	  localNumberOfPreviousElements = numberOfPreviousElements + chunk * nextLevelChunkSize;
	  absolutPos = levelOffset + relativePos;
	  value = bitget(wt, absolutPos);
	  if(value) {
	    relativePos = chunkOffset + nextLevelChunkSize + rank(wtBitmap, absolutPos) - localNumberOfPreviousElements - 1;
	    candidatePositions[newCandidatePositionsSize++] = relativePos;
	  } else if(minc_ < nextLevelChunkUpperLimit) {
	    relativePos += (localNumberOfPreviousElements - rank(wtBitmap, absolutPos));
	    candidatePositions[newCandidatePositionsSize++] = relativePos;
	  }
	}
	candidatePositionsSize = newCandidatePositionsSize;
      }
      
      //subchunks block
      newSubChunkSize = 0;
      for(j = 0; j < subChunkSize; j++) {
	subChunkUpperRelativePos = subChunkUpperLimits[j];
	chunk = subChunkUpperRelativePos / chunkSize;
	chunkOffset = chunk * chunkSize;
	absolutOffset = levelOffset + chunkOffset;
	//numberOfPreviousElements = rank(wtBitmap, absolutOffset)-bitget(wt,absolutOffset);
	localNumberOfPreviousElements = numberOfPreviousElements + chunk * nextLevelChunkSize;
	nextLevelChunk = chunk << 1;
	nextLevelChunkLimit = nextLevelChunk * nextLevelChunkSize;
	nextLevelChunkUpperLimit = nextLevelChunkLimit + nextLevelChunkSize;
	subChunkUpperAbsolutePos = levelOffset + subChunkUpperRelativePos;
	subChunkUpperRank = rank(wtBitmap, subChunkUpperAbsolutePos) - localNumberOfPreviousElements;
	if(minc_ < nextLevelChunkUpperLimit) {
	  //branch 0
	  subChunkUpperRelativePos = 1 + subChunkUpperRelativePos - subChunkUpperRank;
	  if(subChunkUpperRelativePos > nextLevelChunkLimit) {
	    if(subChunkUpperRelativePos == nextLevelChunkUpperLimit  && minc_ <= nextLevelChunkLimit) {
	      for(k = 0; k < nextLevelChunkSize; k++) {
		auxid = ids[nextLevelChunkLimit + k];
		if(bitget(intersector, auxid)) {
		  result[(*resultSize)++] = auxid;
		}
	      }
	      //numReportByLevel[1][i] += nextLevelChunkSize;
	      //results[1] += nextLevelChunkSize;
	    } else {
	      subChunkUpperRelativePos--;
	      if(subChunkUpperRelativePos == 0) {
		candidatePositions[candidatePositionsSize++] = subChunkUpperRelativePos;
	      } else {
		newSubChunkUpperLimits[newSubChunkSize++] = subChunkUpperRelativePos;
	      }
	    }
	  }
	}
	//branch 1 -> always
	subChunkUpperRelativePos = subChunkUpperRank;
	if(subChunkUpperRelativePos > 0) {
	  if(minc_ <= nextLevelChunkUpperLimit && subChunkUpperRelativePos == (nextLevelChunkUpperLimit + nextLevelChunkSize)) {
	    for(k = 0; k < nextLevelChunkSize; k++) {
	      auxid = ids[nextLevelChunkUpperLimit + k];
	      if(bitget(intersector, auxid)) {
		result[(*resultSize)++] = auxid;
	      }
	    }
	    //numReportByLevel[1][i] += nextLevelChunkSize;
	    //results[1] += nextLevelChunkSize;
	  } else {
	    subChunkUpperRelativePos--;
	    if(subChunkUpperRelativePos == 0) {
	      //one single element
	      candidatePositions[candidatePositionsSize++] = chunkOffset + nextLevelChunkSize + subChunkUpperRelativePos;
	    } else {
	      newSubChunkUpperLimits[newSubChunkSize++] = chunkOffset + nextLevelChunkSize + subChunkUpperRelativePos;
	    }
	  }
	}
      }
      subChunkSize = newSubChunkSize;
      
      if(subChunkSize == 0 && candidatePositionsSize == 0) {
	return 0;
      }
      
      elementsInCompleteNodes = (number_of_elements/chunkSize)*chunkSize;
      elementsInNonCompleteNodes = number_of_elements - elementsInCompleteNodes;
      numberOfPreviousElements += elementsInCompleteNodes/2;
      if(elementsInNonCompleteNodes>nextLevelChunkSize)
	numberOfPreviousElements += (elementsInNonCompleteNodes-nextLevelChunkSize);

      chunkSize = nextLevelChunkSize;
      nextLevelChunkSize = nextLevelChunkSize >> 1;
      aux = subChunkUpperLimits;
      subChunkUpperLimits = newSubChunkUpperLimits;
      newSubChunkUpperLimits = aux;
      levelOffset += number_of_elements;
    }
    
    
    for(j = 0; j < candidatePositionsSize; j++) {
      auxid = ids[candidatePositions[j]];
      if(bitget(intersector, auxid)) {
	result[(*resultSize)++] = auxid;
      }
      //numReportByLevel[1][wtLevels-1]++;
      //results[1] ++;
    }
  }
  
  return 0;
}

//Aux functions
int rdt_wt::translateQuery(MBR *mbr, uint *maxax, uint *mincx, uint *maxay, uint *mincy) {
  int auxmaxax = fax->searchRight(mbr->point2.lon);
  int auxmincx = fcx->searchLeft(mbr->point1.lon);
  int auxmaxay = fay->searchRight(mbr->point2.lat);
  int auxmincy = fcy->searchLeft(mbr->point1.lat);
  if(auxmaxax == 0 || auxmincx > (int)number_of_elements || auxmaxay == 0 || auxmincy > (int)number_of_elements) return -1;
  
  *maxax = auxmaxax - 1;
  *mincx = auxmincx - 1;
  *maxay = auxmaxay - 1;
  *mincy = auxmincy - 1;
  
  return 0;
}


/******************************************************************************************
Total memory used by the structure
*******************************************************************************************/
uint rdt_wt::memoryUsage() {
  //printf("%u %u %u %u 2*%d 2*%d %u %u %d\n",fax->memoryUsage(), fay->memoryUsage(), fcx->memoryUsage(), 
	 //fcy->memoryUsage(),number_of_elements*sizeof(uint),(int)wtSize/8,lonwtBitmap->mem_usage, latwtBitmap->mem_usage,sizeof(rdt_wt));
   printf("%u %u %u\n", fax->memoryUsage() + fay->memoryUsage() + fcx->memoryUsage() + fcy->memoryUsage(),
 		      2*number_of_elements*sizeof(uint),
 		      2*wtSize/8 + lonwtBitmap->mem_usage + latwtBitmap->mem_usage+sizeof(rdt_wt));
  return fax->memoryUsage() + fay->memoryUsage() + fcx->memoryUsage() + fcy->memoryUsage() //coords
  + 2*number_of_elements*sizeof(uint) //ids
  + 2*wtSize/8 //wts
  + lonwtBitmap->mem_usage + latwtBitmap->mem_usage; //rank bitmaps
  //+ sizeof(rdt_wt);
}


/******************************************************************************************
Destructs the spatial wavelt tree and returns the memory to the system
*******************************************************************************************/
rdt_wt::~rdt_wt() {
  delete fax;
  delete fay;
  delete fcx;
  delete fcy;
  
  free(idsCX);
  free(idsCY);
  
  destroyBitmap(lonwtBitmap);
  destroyBitmap(latwtBitmap);
  
  free(lonwt);
  free(latwt);
  
  free(subChunkUpperLimits);
  free(newSubChunkUpperLimits);
  free(candidatePositions);
  free(intersector);
}



/******************************************************************************************
Unit test.
*******************************************************************************************/


int main(int argc, char** argv)
{
  if (argc < 11)
  {
    printf("Usage: %s input_file file_type [IdX1Y1X2Y2 | X1Y1X2Y2 | ...] output_file query_input_file1 [query_input_filek] query_file_type [IdX1Y1X2Y2 | X1Y1X2Y2 | ...] query_result encoder blockSize [scaleFactor|1] [test|time].\n", argv[0]);
    return -1;
  }
  
  uint fileType = 0;
  uint queryFileType = 0;
  
  if (strcmp(argv[2], "IdX1Y1X2Y2") == 0) fileType = 0;
  else if (strcmp(argv[2], "X1Y1X2Y2") == 0) fileType = 1;
  else
  {
    printf("Unknown file type.");
    return -1;
  }
  if (strcmp(argv[argc - 6], "IdX1Y1X2Y2") == 0) queryFileType = 0;
  else if (strcmp(argv[argc - 6], "X1Y1X2Y2") == 0) queryFileType = 1;
  else
  {
    printf("Unknown query file type.");
    return -1;
  }
  
  FILE* outputFile = fopen(argv[3], "a");
  
  //read input elements
  uint number_of_input_elements;
  INPUT_MBR *input_data;
  if(fileType == 0) {
    readIdX1Y1X2Y2(argv[1], &input_data, &number_of_input_elements);
  } else if(fileType == 1) {
    readX1Y1X2Y2(argv[1], &input_data, &number_of_input_elements);
  }
  
  //printf("Creando estructuras..\n");
  rdt_wt *theWT = new rdt_wt(input_data, number_of_input_elements, atoi(argv[argc - 3]), argv[argc - 4], atoi(argv[argc - 2]));
  //printf("Estructuras creadas..\n");
  printf("%u;\n",theWT->memoryUsage());
  
  //read result template
  bool haveResultsTemplate = false;
  uint *resultsTemplate;
  uint number_of_results;
  if(strcmp(argv[argc - 5], "no") != 0) {
    readResultsTemplate(argv[argc - 5], &resultsTemplate, &number_of_results);
    haveResultsTemplate = true;
  }
  uint resultsCounter = 0;
  
  //false postives variables
  uint fpCounter = 0;
  
  long* resultWT = (long*) malloc(theWT->getNumberOfElements() * sizeof(long));
  int k;
  
  for(k = 4; k < argc - 6; k++) {
    //read input queries
    uint number_of_input_queries;
    INPUT_MBR *input_queries;
    if(queryFileType == 0) {
      readIdX1Y1X2Y2(argv[k], &input_queries, &number_of_input_queries);
    } else if(queryFileType == 1) {
      readX1Y1X2Y2(argv[k], &input_queries, &number_of_input_queries);
    }
    
    uint i,j;
    uint resultSizeWT;
    
    double time = 0;
    
    if(!strcmp(argv[argc - 1], "time")) {
      fpCounter = 0;
      for(i = 0; i < number_of_input_queries; i++) {
	
	
	tTiempo mitiempo1;
	tTiempo mitiempo2;
	getTiempo(&mitiempo1);
	
	for(j = 0; j < 5; j++) {
	  resultSizeWT = 0;
	  theWT->executeQueryWT(&input_queries[i], resultWT, &resultSizeWT);
	}
	
	getTiempo(&mitiempo2);
	time += ((mitiempo2.user - mitiempo1.user) / 5);
	//fprintf(outputFile, "%ld;%ld;%lf\n", input_queries[i].id, resultSizeWT, (mitiempo2.user - mitiempo1.user) / 10);
	
	if(haveResultsTemplate) {
	  fpCounter += (resultSizeWT - resultsTemplate[resultsCounter++]);
	}
      }
      if(haveResultsTemplate) {	
	fprintf(outputFile, "%lf;%u;%'.4f;", time / number_of_input_queries, fpCounter, 1.0 * fpCounter / number_of_input_queries);
      } else {
	fprintf(outputFile, "%lf;", time / number_of_input_queries);
      }
    } else if(!strcmp(argv[argc - 1], "test")) {
      //long* result = (long*) malloc(number_of_elements * sizeof(long));
      //uint resultSize;
      uint error = 0;
      
      for(i = 0; i < number_of_input_queries; i++) {
	//resultSize = 0;
	resultSizeWT = 0;
	theWT->executeQueryWT(&input_queries[i], resultWT, &resultSizeWT);
	printf("Query %ld: %u elements: ", input_queries[i].id, resultSizeWT); 
	qsort(resultWT, resultSizeWT, sizeof(long), cmplong);
	
	//totalResults += resultSizeWT;
	
	for(j = 0; j < resultSizeWT; j++) {
	  printf("%ld; ", resultWT[j]);
	}
	printf("\n");
	
	  }

	    
	    if(!error) {
	      fprintf(outputFile, "All tests OK.\n");
	    }
      } else {
	printf("Unknown option: %s\n", argv[argc - 1]);
      }
      
      free(input_queries);
    }
    
    if(haveResultsTemplate) {
      free(resultsTemplate);
    }
    free(resultWT);
    
    
    if(!strcmp(argv[argc - 1], "time")) {
      fprintf(outputFile, "%u;", theWT->memoryUsage());
    }
    
    fprintf(outputFile, "\n");
    fclose(outputFile);
    delete theWT;
    
    return 0;
  }


  
